@@ -30,7 +30,7 @@
  */
 
 /*
- * $Id: mailstream_ssl.c,v 1.63 2008/02/17 13:13:26 hoa Exp $
+ * $Id: mailstream_ssl.c,v 1.65 2008/07/04 08:38:01 hoa Exp $
  */
 
 /*
@@ -91,6 +91,23 @@
 
 #include "mailstream_cancel.h"
 
+struct mailstream_ssl_context
+{
+  int fd;
+#ifdef USE_SSL
+#ifndef USE_GNUTLS
+  SSL_CTX * openssl_ssl_ctx;
+  X509* client_x509;
+  EVP_PKEY *client_pkey;
+#else
+  gnutls_session session;
+  gnutls_x509_crt client_x509;
+  gnutls_x509_privkey client_pkey;
+  gnutls_certificate_credentials_t gnutls_credentials;
+#endif
+#endif
+};
+
 #ifdef USE_SSL
 #ifndef USE_GNUTLS
 struct mailstream_ssl_data {
@@ -99,6 +116,7 @@ struct mailstream_ssl_data {
   SSL_CTX * ssl_ctx;
   struct mailstream_cancel * cancel;
 };
+
 #else
 struct mailstream_ssl_data {
   int fd;
@@ -168,6 +186,9 @@ static inline void mailstream_ssl_init(void)
 #ifndef USE_GNUTLS
   if (!openssl_init_done) {
     SSL_library_init();
+    OpenSSL_add_all_digests();
+    OpenSSL_add_all_algorithms();
+    OpenSSL_add_all_ciphers();
     openssl_init_done = 1;
   }
 #else
@@ -254,6 +275,26 @@ mailstream_low_driver * mailstream_ssl_driver = &local_mailstream_ssl_driver;
 static struct mailstream_ssl_context * mailstream_ssl_context_new(SSL_CTX * open_ssl_ctx, int fd);
 static void mailstream_ssl_context_free(struct mailstream_ssl_context * ssl_ctx);
 
+static int mailstream_openssl_client_cert_cb(SSL *ssl, X509 **x509, EVP_PKEY **pkey)
+{
+	struct mailstream_ssl_context * ssl_context = (struct mailstream_ssl_context *)SSL_CTX_get_app_data(ssl->ctx);
+	
+	if (x509 == NULL || pkey == NULL) {
+		return 0;
+	}
+
+	if (ssl_context == NULL)
+		return 0;
+
+	*x509 = ssl_context->client_x509;
+	*pkey = ssl_context->client_pkey;
+
+	if (*x509 && *pkey)
+		return 1;
+	else
+		return 0;
+}
+
 static struct mailstream_ssl_data * ssl_data_new_full(int fd, SSL_METHOD * method, void (* callback)(struct mailstream_ssl_context * ssl_context, void * cb_data), void * cb_data)
 {
   struct mailstream_ssl_data * ssl_data;
@@ -261,6 +302,7 @@ static struct mailstream_ssl_data * ssl_data_new_full(int fd, SSL_METHOD * metho
   int r;
   SSL_CTX * tmp_ctx;
   struct mailstream_cancel * cancel;
+  struct mailstream_ssl_context * ssl_context = NULL;
   
   mailstream_ssl_init();
   
@@ -269,13 +311,12 @@ static struct mailstream_ssl_data * ssl_data_new_full(int fd, SSL_METHOD * metho
     goto err;
   
   if (callback != NULL) {
-    struct mailstream_ssl_context * ssl_context;
-    
     ssl_context = mailstream_ssl_context_new(tmp_ctx, fd);
     callback(ssl_context, cb_data);
-    mailstream_ssl_context_free(ssl_context);
   }
   
+  SSL_CTX_set_app_data(tmp_ctx, ssl_context);
+  SSL_CTX_set_client_cert_cb(tmp_ctx, mailstream_openssl_client_cert_cb);
   ssl_conn = (SSL *) SSL_new(tmp_ctx);
   if (ssl_conn == NULL)
     goto free_ctx;
@@ -321,6 +362,7 @@ again:
   ssl_data->ssl_conn = ssl_conn;
   ssl_data->ssl_ctx = tmp_ctx;
   ssl_data->cancel = cancel;
+  mailstream_ssl_context_free(ssl_context);
 
   return ssl_data;
 
@@ -330,6 +372,7 @@ again:
   SSL_free(ssl_conn);
  free_ctx:
   SSL_CTX_free(tmp_ctx);
+  mailstream_ssl_context_free(ssl_context);
  err:
   return NULL;
 }
@@ -345,6 +388,29 @@ static struct mailstream_ssl_data * tls_data_new(int fd, void (* callback)(struc
 }
 
 #else
+
+static struct mailstream_ssl_context * mailstream_ssl_context_new(gnutls_session session, int fd);
+static void mailstream_ssl_context_free(struct mailstream_ssl_context * ssl_ctx);
+
+static int mailstream_gnutls_client_cert_cb(gnutls_session session,
+                               const gnutls_datum *req_ca_rdn, int nreqs,
+                               const gnutls_pk_algorithm *sign_algos,
+                               int sign_algos_length, gnutls_retr_st *st)
+{
+	struct mailstream_ssl_context * ssl_context = (struct mailstream_ssl_context *)gnutls_session_get_ptr(session);
+	gnutls_certificate_type type = gnutls_certificate_type_get(session);
+	st->ncerts = 0;
+	if (type == GNUTLS_CRT_X509 && ssl_context->client_x509 && ssl_context->client_pkey) {
+		st->ncerts = 1;
+		st->type = type;
+		st->cert.x509 = &(ssl_context->client_x509);
+		st->key.x509 = ssl_context->client_pkey;
+		st->deinit_all = 0;
+		return 0;
+	}
+	return -1;
+}
+
 static struct mailstream_ssl_data * ssl_data_new(int fd, void (* callback)(struct mailstream_ssl_context * ssl_context, void * cb_data), void * cb_data)
 {
   struct mailstream_ssl_data * ssl_data;
@@ -365,6 +431,7 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, void (* callback)(struc
 
   gnutls_certificate_credentials_t xcred;
   int r;
+  struct mailstream_ssl_context * ssl_context = NULL;
   
   mailstream_ssl_init();
   
@@ -375,14 +442,23 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, void (* callback)(struc
   if (session == NULL || r != 0)
     return NULL;
   
+  if (callback != NULL) {
+    ssl_context = mailstream_ssl_context_new(session, fd);
+    callback(ssl_context, cb_data);
+  }
+  
+  gnutls_session_set_ptr(session, ssl_context);
+  gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
+  gnutls_certificate_client_set_retrieve_function(xcred, mailstream_gnutls_client_cert_cb);
+
   gnutls_set_default_priority(session);
   gnutls_protocol_set_priority (session, proto_prio);
   gnutls_cipher_set_priority (session, cipher_prio);
   gnutls_kx_set_priority (session, kx_prio);
   gnutls_mac_set_priority (session, mac_prio);
 
-  gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
   gnutls_dh_set_prime_bits(session, 512);
+
   gnutls_transport_set_ptr(session, (gnutls_transport_ptr) fd);
 
   /* lower limits on server key length restriction */
@@ -393,6 +469,7 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, void (* callback)(struc
   } while (r == GNUTLS_E_AGAIN || r == GNUTLS_E_INTERRUPTED);
 
   if (r < 0) {
+    gnutls_perror(r);
     goto free_ssl_conn;
   }
   
@@ -413,12 +490,15 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, void (* callback)(struc
   ssl_data->xcred = xcred;
   ssl_data->cancel = cancel;
   
+  mailstream_ssl_context_free(ssl_context);
+
   return ssl_data;
   
  free_cancel:
   mailstream_cancel_free(cancel);
  free_ssl_conn:
   gnutls_certificate_free_credentials(xcred);
+  mailstream_ssl_context_free(ssl_context);
   gnutls_deinit(session);
  err:
   return NULL;
@@ -692,7 +772,7 @@ static int wait_write(mailstream_low * s)
   if (fd > max_fd)
     max_fd = fd;
   
-  r = select(ssl_data->fd + 1, &fds_read, &fds_write, NULL, &timeout);
+  r = select(max_fd + 1, &fds_read, &fds_write, NULL, &timeout);
   if (r == 0)
     return -1;
   
@@ -882,19 +962,6 @@ static void mailstream_low_ssl_cancel(mailstream_low * s)
 #endif
 }
 
-struct mailstream_ssl_context
-{
-  int fd;
-#ifdef USE_SSL
-#ifndef USE_GNUTLS
-  SSL_CTX * openssl_ssl_ctx;
-#else
-  gnutls_session gnutls_session;
-  gnutls_certificate_credentials_t gnutls_credentials;
-#endif
-#endif
-};
-
 mailstream_low * mailstream_low_ssl_open_with_callback(int fd,
     void (* callback)(struct mailstream_ssl_context * ssl_context, void * data), void * data)
 {
@@ -932,6 +999,66 @@ int mailstream_ssl_set_client_certicate(struct mailstream_ssl_context * ssl_cont
 #endif /* USE_SSL */
 }
 
+LIBETPAN_EXPORT
+int mailstream_ssl_set_client_certificate_data(struct mailstream_ssl_context * ssl_context,
+    unsigned char *x509_der, size_t len)
+{
+#ifdef USE_SSL
+#ifndef USE_GNUTLS
+  X509 *x509 = NULL;
+  if (x509_der != NULL && len > 0)
+    x509 = d2i_X509(NULL, (const unsigned char **)&x509_der, len);
+  ssl_context->client_x509 = (X509 *)x509;
+  return 0;
+#else
+  gnutls_datum tmp;
+  int r;
+  ssl_context->client_x509 = NULL;
+  if (len == 0)
+    return 0;
+  gnutls_x509_crt_init(&(ssl_context->client_x509));
+  tmp.data = x509_der;
+  tmp.size = len;
+  if ((r = gnutls_x509_crt_import(ssl_context->client_x509, &tmp, GNUTLS_X509_FMT_DER)) < 0) {
+    gnutls_x509_crt_deinit(ssl_context->client_x509); /* ici */
+    ssl_context->client_x509 = NULL;
+    return -1;
+  }
+  return 0;
+#endif
+#endif
+  return -1;
+}
+int mailstream_ssl_set_client_private_key_data(struct mailstream_ssl_context * ssl_context,
+    unsigned char *pkey_der, size_t len)
+{
+#ifdef USE_SSL
+#ifndef USE_GNUTLS
+  EVP_PKEY *pkey = NULL;
+  if (pkey_der != NULL && len > 0)
+    pkey = d2i_AutoPrivateKey(NULL, (const unsigned char **)&pkey_der, len);
+  ssl_context->client_pkey = (EVP_PKEY *)pkey;
+  return 0;
+#else
+  gnutls_datum tmp;
+  int r;
+  ssl_context->client_pkey = NULL;
+  if (len == 0)
+    return 0;
+  gnutls_x509_privkey_init(&(ssl_context->client_pkey));
+  tmp.data = pkey_der;
+  tmp.size = len;
+  if ((r = gnutls_x509_privkey_import(ssl_context->client_pkey, &tmp, GNUTLS_X509_FMT_DER)) < 0) {
+    gnutls_x509_crt_deinit(ssl_context->client_pkey);
+    ssl_context->client_pkey = NULL;
+    return -1;
+  }
+  return 0;
+#endif
+#endif
+  return -1;
+}
+
 int mailstream_ssl_set_server_certicate(struct mailstream_ssl_context * ssl_context, 
     char * CAfile, char * CApath)
 {
@@ -963,6 +1090,8 @@ static struct mailstream_ssl_context * mailstream_ssl_context_new(SSL_CTX * open
     return NULL;
   
   ssl_ctx->openssl_ssl_ctx = open_ssl_ctx;
+  ssl_ctx->client_x509 = NULL;
+  ssl_ctx->client_pkey = NULL;
   ssl_ctx->fd = fd;
   
   return ssl_ctx;
@@ -972,6 +1101,31 @@ static void mailstream_ssl_context_free(struct mailstream_ssl_context * ssl_ctx)
 {
   free(ssl_ctx);
 }
+#else
+static struct mailstream_ssl_context * mailstream_ssl_context_new(gnutls_session session, int fd)
+{
+  struct mailstream_ssl_context * ssl_ctx;
+  
+  ssl_ctx = malloc(sizeof(* ssl_ctx));
+  if (ssl_ctx == NULL)
+    return NULL;
+  
+  ssl_ctx->session = session;
+  ssl_ctx->client_x509 = NULL;
+  ssl_ctx->client_pkey = NULL;
+  ssl_ctx->fd = fd;
+  
+  return ssl_ctx;
+}
+
+static void mailstream_ssl_context_free(struct mailstream_ssl_context * ssl_ctx)
+{
+  if (ssl_ctx->client_x509)
+    gnutls_x509_crt_deinit(ssl_ctx->client_x509);
+  if (ssl_ctx->client_pkey)
+    gnutls_x509_privkey_deinit(ssl_ctx->client_pkey);
+  free(ssl_ctx);
+}
 #endif
 #endif
 
